/*
 * BEGIN FLOCK GPL
 *
 * Copyright Flock Inc. 2005-2008
 * http://flock.com
 *
 * This file may be used under the terms of the
 * GNU General Public License Version 2 or later (the "GPL"),
 * http://www.gnu.org/licenses/gpl.html
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * END FLOCK GPL
 */

const CC = Components.classes;
const CI = Components.interfaces;

const KEEPACCOUNT_TIMER = 5000;
const REBUILD_ACCOUNT_DELAY = 3000;

const SERVICE_SHOW_PREF_BRANCH = "flock.notification.show.accounts";
const SERVICE_SHOW_PREF_HEADER = ".account-setup-info-";

const IS_AUTHENTICATED_RSC = "http://flock.com/rdf#isAuthenticated";
const IS_TRANSIENT_RSC = "http://flock.com/rdf#isTransient";
const AUTH_TOKEN_RSC = "http://flock.com/rdf#authtoken";
const ACCOUNTS_ROOT_RSC = "http://flock.com/rdf#authtoken";

const BROWSER_HEIGHT_OFFSET = 50;

var accountsSidebar =  {
  sidebarObserver : null,
  gAccountsArea : ["people", "media", "bookmarks", "manualBlog", "webmail"],
  _metrics : CC["@flock.com/metrics-service;1"]
             .getService(CI.flockIMetricsService),
  services : null,
  prefService : CC["@mozilla.org/preferences-service;1"]
                .getService(CI.nsIPrefBranch),
  prefBranch : CC["@mozilla.org/preferences-service;1"]
               .getService(CI.nsIPrefService)
               .getBranch(SERVICE_SHOW_PREF_BRANCH),
  acUtils : CC["@flock.com/account-utils;1"]
            .getService(CI.flockIAccountUtils),
  
  /*
   Observer for the AccountsRoot rdf
   */
  onAssert : function as_onAssert(aDatasource, aSource, aPredicate, aTarget) {
    if (aPredicate.Value == IS_TRANSIENT_RSC ||
        aPredicate.Value == AUTH_TOKEN_RSC)
    {
      accountsSidebar.rebuildAccounts();
    }
  },
  
  onUnassert : function as_onUnassert(aDatasource, aSource, aPredicate,
                                      aTarget)
  {
    if (aSource.Value == ACCOUNTS_ROOT_RSC) {
      accountsSidebar.rebuildAccounts();
    }
  },
  
  onChange : function as_onChange(aDatasource, aSource, aPredicate, aOldTarget,
                                  aNewTarget)
  {
    if (aPredicate.Value == IS_AUTHENTICATED_RSC) {
      accountsSidebar.rebuildAccounts();
    } else if (aPredicate.Value == IS_TRANSIENT_RSC) {
      if (aNewTarget instanceof CI.nsIRDFLiteral) {
        // If we are "keeping" this account then delay the account refresh to
        // prevent user double clicking and forgetting the account.
        if (aNewTarget.Value == "false") {
          // Delay rebuilding accounts so user does not double click.
          setTimeout("accountsSidebar.rebuildAccounts()", REBUILD_ACCOUNT_DELAY);
        } else {
          accountsSidebar.rebuildAccounts();
        }
      }
    }
  },
  
  onMove : function as_onMove(aDatasource, aOldSource, aNewSource, aPredicate,
                              aTarget)
  {
  },
  
  onBeginUpdateBatch : function as_onBeginUpdateBatch(aDatasource) {
  },
  
  onEndUpdateBatch : function as_onEndUpdateBatch(aDatasource) {
  },

  /*
   Preference observer for the "flock.notification.show." branch
   */
  observe : function as_observe(aSubject, aTopic, aData) {
    // Observe commands to select a service when opening
    if (aTopic == "accountsSidebar") {
      var availableServiceList = $("availableServicesList");
      var folderList = availableServiceList.children;
      for (var iFolderCount = 0; iFolderCount < folderList.length; iFolderCount++) {
        var foundItem = false;
        var mFolder = document.getAnonymousElementByAttribute(folderList[iFolderCount],
                                                              "anonid", "title");
        if (mFolder.value == aData) {
          // They just want the folder open with no particular service selected.
          foundItem = true;
        } else {
          // Check if they want a service in this groupsvn 
          var serviceList = folderList[iFolderCount].children;
          for (var iServiceCount = 0; iServiceCount < serviceList.length && !foundItem; iServiceCount++) {
            if (serviceList[iServiceCount].value == aData) {
              // We found a matching service so select and ensure that this is visible
              availableServiceList.clearSelection();
              serviceList[iServiceCount].selected = "true";
              availableServiceList.ensureElementIsVisible(serviceList[iServiceCount]);
              foundItem = true;
            }
          }
        }
        // Since the service we want is not in this area we need to close the folder
        // this way only the folder that has the service we want to highlight is open
        if (!foundItem) {
          folderList[iFolderCount].setAttribute("open", "false");
          folderList[iFolderCount].toggleOpen(null); 
        } else {
          folderList[iFolderCount].setAttribute("open", "true");
        }
      }
      return;
    }
  
    // We can make this better by actually moving the child elements arround
    // rather than redrawing the whole tree again.
    if (aTopic == "nsPref:changed") {
      // data will be the "show.account-setup-info-" + svc.shortName
      this.loadAvailableServices();
    }
    
    if (aTopic == "toggleAccountsArea") {
      this.openAccountArea(aData);
    }
  },

  QueryInterface : function (aIID) {
    if ( !aIID.equals(Components.interfaces.nsISupports) &&
          !aIID.equals(Components.interfaces.nsIObserver) )
    {
      throw Components.results.NS_ERROR_NO_INTERFACE;
    }
    return this;
  }
}

/* Rebuild all accounts in sidebar.
 */
accountsSidebar.rebuildAccounts =
function as_rebuildAccounts() {
  $("activeAccountsList").builder.rebuild();
  this.toggleEnabledAccounts();
}

accountsSidebar.toggleEnabledAccounts =
function as_toggleEnabledAccounts() {
  if ($("activeAccountsList").getRowCount() > 0) {
    $("activeAccountsSectionDeck").selectedIndex = 0;
  } else {
    $("activeAccountsSectionDeck").selectedIndex = 1;
  }
}

//! Initialize the sidebar controls for Accounts and Services.
/*!
  This function will be called when the Accounts and Services sidebar is loaded.
  It will load and group the available services in the Available Services tree control.
*/
accountsSidebar.onLoad =
function as_onLoad() {
  this._metrics.report("AccountSidebar-Opened", null);

  this.prefBranch.QueryInterface(Components.interfaces.nsIPrefBranch2);
  this.prefBranch.addObserver("", this, false);
                         
  this._logger = Components.classes['@flock.com/logger;1'].createInstance(Components.interfaces.flockILogger);
  this._logger.init('accountsSidebar');
  this._logger.info('starting up...');

  // Bug 11962: Set a max height so that when enough configured accounts are
  // added the available accounts and services aren't pushed off the screen
  var windowHeight = window.innerHeight;
  if (windowHeight) {
    var maxHeight = windowHeight - BROWSER_HEIGHT_OFFSET;
    document.getElementById("activeAccountsSectionDeck")
            .setAttribute("maxheight", maxHeight);
  }

  // Load the coop so we can use it later.
  this.coop = Components.classes["@flock.com/singleton;1"]
                        .getService(Components.interfaces.flockISingleton)
                        .getSingleton("chrome://flock/content/common/load-faves-coop.js")
                        .wrappedJSObject;

  // So we can be localized
  this.strbundle = document.getElementById("accountStrings");

  // Get all the available services
  this.services = accountsSidebar.getFlockWebServices();

  // Load all available services
  this.loadAvailableServices();

  this.sidebarObserver = Components.classes["@mozilla.org/observer-service;1"]
                        .getService(Components.interfaces.nsIObserverService);
  this.sidebarObserver.addObserver(this, "accountsSidebar", false);
  this.sidebarObserver.addObserver(this, "toggleAccountsArea", false);
  // Set up observer for the flock-favorites
  var rdfService = Components.classes["@mozilla.org/rdf/rdf-service;1"]
                             .getService(Components.interfaces.nsIRDFService);

  var accountsDS = rdfService.GetDataSource("rdf:flock-favorites");
  accountsDS.AddObserver(accountsSidebar);
                            
  this.wsMgr = Components.classes["@flock.com/webservice-manager;1"]
                          .getService(Components.interfaces.flockIWebServiceManager);
  this.toggleEnabledAccounts();
}

//! Shutdown the sidebar controls for Accounts and Services.
/*!
  This function will be called when the Accounts and Services sidebar is unloaded.
  It will unregister observers.
*/
accountsSidebar.onUnLoad =
function as_onUnload() {
  if (!this.sidebarObserver) {
    // not even loaded.
    return;
  }
  this._logger.info('onUnload');
  this._metrics.report("AccountSidebar-Close", null);

  var rdfService = Components.classes["@mozilla.org/rdf/rdf-service;1"]
                             .getService(Components.interfaces.nsIRDFService);

  var accountsDS = rdfService.GetDataSource("rdf:flock-favorites");
  accountsDS.RemoveObserver(accountsSidebar);

  this.prefBranch.removeObserver("", this, false);

  this.sidebarObserver.removeObserver(this, "accountsSidebar");
  this.sidebarObserver.removeObserver(this, "toggleAccountsArea");
}

accountsSidebar.showPopup =
function as_showPopup(data) {
  var popup = $("tipcontainer");
 
  // Remove old children 
  while(popup.firstChild) {
    popup.removeChild(popup.firstChild);
  }

  var text = document.createTextNode(data);
  var desc = document.createElement("description");
  desc.appendChild(text);
  popup.appendChild(desc);
  
  popup.setAttribute("height", desc.boxObject.height + 5);
}
//! Loads all available services.
/*!
  Load all available services in to the Available Services tree and group
  them by service type.
    Blogging
*/
accountsSidebar.loadAvailableServices =
function as_loadAvailableServices() {
  // This should be turned into a xbl...
  // Populates the active service list
  var availableServiceList = $("availableServicesList");

  // Setup our groups
  var groupServiceArray = [];
  groupServiceArray["People"] = [];
  groupServiceArray["Media"] = [];
  groupServiceArray["Blog"] = [];
  groupServiceArray["Favorites"] = [];
  groupServiceArray["Webmail"] = [];

  // Clean the account list
  while (availableServiceList.getRowCount() > 0) {
    availableServiceList.removeItemAt(0);
  }

  // Loop through services to catagorize them.
  for (var i = 0; i < this.services.length; i++) {
    var contractID = this.services[i][0];
    var svc = this.services[i][1];

    var accountitem = document.createElement("richtreeitem");
    var accountLabel = document.createElement("label");
    var accountFavIcon = document.createElement("image");
    var accountBox = document.createElement("hbox");
    var tipData;

    // <richtreeitem value="contractID">
    //     <hbox onclick="">
    //       <image src="svc.icon" />
    //       <label value="Flickr" flex="1" tooltiptext="&flock.services.accountsSidebar.tooltips.XXXXX;" />
    //     </hbox>
    // </richtreeitem>

    accountitem.setAttribute('value', contractID);
    accountBox.setAttribute('flex', '1');
    accountFavIcon.setAttribute("src", svc.icon);
    accountLabel.setAttribute('value', svc.title);
    accountLabel.setAttribute('flex', '1');
    //accountLabel.setAttribute('onclick', 'accountsSidebar.gotoServiceLoginPage();');

    accountBox.appendChild(accountFavIcon);
    accountBox.appendChild(accountLabel);
    accountitem.appendChild(accountBox);

    if (svc.shortName != null) {
      // Set up the tooltip data
      accountitem.setAttribute('tooltip','aas-tooltip');
      try {
        tipData = this.strbundle.getString("flock.services.accountsSidebar.tooltips." + svc.shortName)
      } catch (err) {
        tipData = this.strbundle.getString("flock.services.accountsSidebar.tooltips.default");
      }
      accountitem.setAttribute('onmouseover', "accountsSidebar.showPopup(\""+tipData+"\")");
    }

			  
    // Note: Order of services checking here is important! If a service implements
    // more than 1 component (Ie: Blog and Photo) then the first one checked here
    // is where it will be catagorized.

    if (svc instanceof Components.interfaces.flockISocialWebService) {
      groupServiceArray["People"].push(accountitem);
      continue;
    }

    // Try and put it into the Blogging catagory
    // (Blogger, Blogsome, LiveJournal, Typepad, WordPress, Xanga...
    if (svc instanceof Components.interfaces.flockIBlogWebService) {
      groupServiceArray["Blog"].push(accountitem);
      continue;
    }

    // Next try the Media catagory (Flickr, Photobucket, YouTube...
    if (svc instanceof Components.interfaces.flockIMediaWebService) {
      groupServiceArray["Media"].push(accountitem);
      continue;
    }

    // Next try the bookmark catagory (del.icio.us, Magnolia, Shadows...)
    if (svc instanceof Components.interfaces.flockIBookmarkWebService) {
      groupServiceArray["Favorites"].push(accountitem);
      continue;
    }

    // Next try the webmail catagory (Yahoo! Mail, Gmail, MSN Hotmail...)
    if (svc instanceof Components.interfaces.flockIMailWebService) {
      groupServiceArray["Webmail"].push(accountitem);
      continue;
    }

  }

///////// Here is where we order the services into groups /////////

  // Social ("People") web services
  var accountFolder = document.createElement("richtreefolder");
  accountFolder.setAttribute("title",
                             this.strbundle.getString("flock.services.accountsSidebar.groups.people"));
  accountFolder.setAttribute("id",
                             "flock.services.accountsSidebar.groups.people");
  accountFolder.setAttribute("open", "true");
  if (groupServiceArray["People"].length > 0) {
    accountFolder.setAttribute("open", "true");
  }
  for (var i = 0; i < groupServiceArray["People"].length; i++) {
    accountFolder.appendChild(groupServiceArray["People"][i]);
  }
  availableServiceList.appendChild(accountFolder);

///////// BEGIN Media     //////////////
  var accountFolder = document.createElement("richtreefolder");
  accountFolder.setAttribute('title', this.strbundle.getString("flock.services.accountsSidebar.groups.media"));
  accountFolder.setAttribute('id', "flock.services.accountsSidebar.groups.media");
  accountFolder.setAttribute('open', 'true');
  if (groupServiceArray["Media"].length > 0) {
    accountFolder.setAttribute('open', 'true');
  }
  for (var i = 0; i < groupServiceArray["Media"].length; i++) {
    accountFolder.appendChild(groupServiceArray["Media"][i]);
  }
  availableServiceList.appendChild(accountFolder);
///////// END Media       //////////////

///////// BEGIN BLOGS //////////////
  var accountFolder = document.createElement("richtreefolder");
  accountFolder.setAttribute('title', this.strbundle.getString("flock.services.accountsSidebar.groups.blog"));
  accountFolder.setAttribute('id', "flock.services.accountsSidebar.groups.blog");
  accountFolder.setAttribute('open', 'true');
  if (groupServiceArray["Blog"].length > 0) {
    accountFolder.setAttribute('open', 'true');
  }
  for (var i = 0; i < groupServiceArray["Blog"].length; i++) {
    accountFolder.appendChild(groupServiceArray["Blog"][i]);
  }

  var accountitem = document.createElement("richtreeitem");
  var accountLabel = document.createElement("label");  
  var accountFavIcon = document.createElement("image");
  var accountBox = document.createElement("hbox");

  // Add the special service for self-hosted blog
  accountitem.setAttribute('value', 'CustomBlog');
  accountBox.setAttribute('flex', '1');
  accountFavIcon.setAttribute("class", "self-hosted-blog");
  accountLabel.setAttribute('value', this.strbundle.getString("flock.services.accountsSidebar.groups.manualBlog"));
  accountFolder.setAttribute('id', "flock.services.accountsSidebar.groups.manualBlog");
  accountLabel.setAttribute('flex', '1');
  accountLabel.setAttribute('tooltiptext', this.strbundle.getString("flock.services.accountsSidebar.tooltips.manualBlog"));
  accountBox.appendChild(accountFavIcon);
  accountBox.appendChild(accountLabel);
  accountitem.appendChild(accountBox);
  accountFolder.appendChild(accountitem);
  availableServiceList.appendChild(accountFolder);
///////// END BLOGS       //////////////

///////// BEGIN Bookmarks //////////////
  var accountFolder = document.createElement("richtreefolder");
  accountFolder.setAttribute('title', this.strbundle.getString("flock.services.accountsSidebar.groups.bookmarks"));
  accountFolder.setAttribute('id', "flock.services.accountsSidebar.groups.bookmarks");
  accountFolder.setAttribute('open', 'true');
  if (groupServiceArray["Favorites"].length > 0) {
    accountFolder.setAttribute('open', 'true');
  }
  for (var i = 0; i < groupServiceArray["Favorites"].length; i++) {
    accountFolder.appendChild(groupServiceArray["Favorites"][i]);
  }
  availableServiceList.appendChild(accountFolder);
///////// END Bookmarks   //////////////

  // Webmail services
  accountFolder = document.createElement("richtreefolder");
  accountFolder.setAttribute("title",
                             this.strbundle.getString("flock.services.accountsSidebar.groups.webmail"));
  accountFolder.setAttribute("id",
                             "flock.services.accountsSidebar.groups.webmail");
  accountFolder.setAttribute("open", "true");
  if (groupServiceArray["Webmail"].length > 0) {
    accountFolder.setAttribute("open", "true");
  }
  for (var i = 0; i < groupServiceArray["Webmail"].length; i++) {
    accountFolder.appendChild(groupServiceArray["Webmail"][i]);
  }
  availableServiceList.appendChild(accountFolder);
}

//! Loads all available services into an array.
/*!
  This will load all available services into an array of contractid and service object.
*/
accountsSidebar.getFlockWebServices =
function as_getFlockWebServices() {
  var topPeople;
  this.prefService.getPrefType("flock.accounts.top.people") && (topPeople =
  this.prefService.getCharPref("flock.accounts.top.people"));

  var topMedia;
  this.prefService.getPrefType("flock.accounts.top.media") && (topMedia =
  this.prefService.getCharPref("flock.accounts.top.media"));
  
  var topBlog;
  this.prefService.getPrefType("flock.accounts.top.blog") && (topBlog =
  this.prefService.getCharPref("flock.accounts.top.blog"));

  var topFavorites;
  this.prefService.getPrefType("flock.accounts.top.favorite") && (topFavorites =
  this.prefService.getCharPref("flock.accounts.top.favorite"));
  
  var topWebmail;
  this.prefService.getPrefType("flock.accounts.top.webmail") && (topWebmail =
  this.prefService.getCharPref("flock.accounts.top.webmail"));
  
  var CATMAN = Components.classes['@mozilla.org/categorymanager;1'].getService(Components.interfaces.nsICategoryManager);
  var webServices = new Array();
  var e = CATMAN.enumerateCategory('flockWebService');
  while (e.hasMoreElements()) {
    try {
      var entry = e.getNext().QueryInterface(Components.interfaces.nsISupportsCString);
      if (!entry) { continue; }
      var contractID = CATMAN.getCategoryEntry('flockWebService', entry.data);
      var svc = Components.classes[contractID].getService(Components.interfaces.flockIWebService);

      // Check to see if we should NOT show this in the list
      if (svc.isHidden()) {
        continue;
      }

      if ((svc.shortName == topPeople) ||
          (svc.shortName == topMedia) ||
          (svc.shortName == topBlog) ||
          (svc.shortName == topFavorites) ||
          (svc.shortName == topWebmail))
      {
        webServices.unshift([contractID, svc]);
      } else {
        webServices.push([contractID, svc]);
      }
    } catch (e) {
      this._logger.info('Error occured in getFlockWebServices: ' + e);
    }
  }
  return webServices;
}

//! Go to a services login page.
/*!
  This will navigate to the defined login page for a service when the user
  double clicks on an entry in the Available services list.
*/
accountsSidebar.gotoServiceLoginPage =
function as_gotoServiceLoginPage(aEvent) {
  if (!$('availableServicesList').selectedItem) {
    return;
  }
  var contractID = $('availableServicesList').selectedItem.value;

  switch (contractID) {
    case 'CustomBlog':
      this._metrics.report("AccountSidebar-ServiceClick", contractID);
      // We need to open the custom blog editor instead
      window.openDialog("chrome://flock/content/blog/blogAccountWizard.xul",
                        "accountWizard",
                        "chrome,modal,centerscreen",
                        { defaultToCustom: true });
      accountsSidebar.rebuildAccounts();
      break;
    case 'ApplyForService':
      // Here we just direct to a special url that the user can
      // request a service to add.
      // NOTE: Put this in a properties file!!! or better yet branding file
      var flockApplyForServiceUrl = "http://www.flock.com/";
      openUILinkIn(flockApplyForServiceUrl, "tab");
      break;
    default:
      var cid = Components.classes[contractID];
      if (!cid) {
        // No cid exists, folder item or item is invalid.
        return;
      }

      var svc = cid.getService(Components.interfaces.flockIWebService);
      this._metrics.report("AccountSidebar-ServiceClick", svc.shortName);
      if (svc) {
        var prefName = SERVICE_SHOW_PREF_BRANCH + SERVICE_SHOW_PREF_HEADER + svc.shortName;
        if (this.prefService.getPrefType(prefName) &&
            !this.prefService.getBoolPref(prefName))
        {
          // Re-enable this service
          this.prefService.setBoolPref(prefName, true);
        }
        // Tell the Web Service Manager to automatically keep any accounts
        // created for this service during the next while...
        this.wsMgr.autoKeepAccountForService(svc.urn);
        
        // Display the login page for the service
        var redirURL;

        if (svc.redirUrl) {
          redirURL = svc.redirUrl;
        } else {
          redirURL = "http://redir.flock.com/redir?component=browser&partner="
                     + svc.shortName + "&action=accountssidebar_click";
        }

        if (aEvent.button != 2) {
          // Left or Middle clicks ok
          var openIn = "current";

          if (aEvent.button == 1) {
            // Middle click
            openIn = "tab";
          }
//@line 602 "/home/build/tinderbuild_blue/src/flock/mozilla/flock/base/services/common/content/accountsSidebar.js"
          if (aEvent.ctrlKey) {
            // Ctrl Key held down
            openIn = "tab";
          }
//@line 607 "/home/build/tinderbuild_blue/src/flock/mozilla/flock/base/services/common/content/accountsSidebar.js"
          if (aEvent.shiftKey) {
            // Shift key held down
            openIn = "window";
          }
          var win = CC["@mozilla.org/appshell/window-mediator;1"]
                    .getService(CI.nsIWindowMediator)
                    .getMostRecentWindow("navigator:browser");
          if (win &&
              (openIn == "current") &&
              (win.gBrowser.currentURI.spec == "about:myworld"))
          {
            // Never load over MyWorld.  Load in new tab
            openIn = "tab";
          }

          win.openUILinkIn(redirURL, openIn);
        }
      }
      break;
    }

}

//! Change Message
/*!
  This will change the message displayed directly above the Available Accounts
  we have it in a function so that if we want we can add timers to return to a
  default message.
*/
accountsSidebar.getNotificationBox =
function as_getNotificationBox() {
  var mainWindow = window.QueryInterface(Components.interfaces.nsIInterfaceRequestor)
                      .getInterface(Components.interfaces.nsIWebNavigation)
                      .QueryInterface(Components.interfaces.nsIDocShellTreeItem)
                      .rootTreeItem
                      .QueryInterface(Components.interfaces.nsIInterfaceRequestor)
                      .getInterface(Components.interfaces.nsIDOMWindow);
  var gBrowser = mainWindow.getBrowser();
  var notificationBox = gBrowser.getNotificationBox();

  return notificationBox;
}

accountsSidebar.clearNotifications =
function as_clearNotifications(aNotificationList) {
  var notificationBox = this.getNotificationBox();

  for (var iIndex = 0; iIndex < aNotificationList.length; iIndex++) {
    var oldNotification = notificationBox.getNotificationWithValue(aNotificationList[iIndex]);
    if (oldNotification) {
      notificationBox.removeNotification(oldNotification);
    }
  }
}


//! Forget Account
/*!
  This will remove the selected account from the browser, prompting the user
  before doing so.
*/
accountsSidebar.forgetAccount =
function as_forgetAccount(aAccountCard, aEvent) {

  // Ignore right mouse click
  if (aEvent.button == 2) {
    return;
  }

  var id = aAccountCard.parentNode.id;
  var account = this.coop.get(id);
  var svc = CC[account.serviceId].getService(CI.flockIWebService);

  var prompts = CC["@mozilla.org/embedcomp/prompt-service;1"]
                .getService(CI.nsIPromptService);

  if (prompts.confirm(window, 
                   this.strbundle.getString("flock.services.accountsSidebar.remove.title") ,
                   this.strbundle.getString("flock.services.accountsSidebar.remove.msg"))) {
    svc.getAccount(id).logout(null);
    svc.removeAccount(id);

    this._metrics.report("Account-Forget", svc.shortName);

    this.sidebarObserver.notifyObservers(null, 'refresh-myworld-accounts', false);
    this.clearNotifications(["account-setup-success-" + svc.shortName,
                             "account-setup-info-" + svc.shortName]);

    var notificationBox = this.getNotificationBox();
    var message = this.strbundle.getFormattedString("flock.services.accountsSidebar.removedAccount",
                                                    [svc.title]);
    notificationBox.appendNotification( message,
      "account-remove-success-" + svc.shortName,
      "chrome://browser/skin/Info.png",
      notificationBox.FLOCK_PRIORITY_MEDIUM, null, null );
  }
}

//! Keep Account
/*!
  When a Transient account is selected this will allow the user to keep said
  account after restarts of the browser.
*/
accountsSidebar.keepAccount =
function as_keepAccount(aKeepButton) {
  // Disable the button, to prevent double clicks and forgetting the account.
  aKeepButton.setAttribute("disabled", true);
  
  // Get the account card
  var aAccountCard = document.getBindingParent(aKeepButton);
  
  var SB = "common/notificationOverlay";
  if (aAccountCard.parentNode.getAttribute("keepingaccount") == "true") {
    // We are in the process of keeping this account so don't allow for double clicks
    return;
  }

  var id = aAccountCard.parentNode.id;
  var account = this.coop.get(id);
  var svc = Components.classes[account.serviceId].getService(Components.interfaces.flockIWebService);
  svc.getAccount(id).keep();
  var message = null;
  var link = null;
  var appName = brandGetString("brand", "brandShortName");
  
  if(svc instanceof Components.interfaces.flockIBookmarkWebService)
  {
    link = [{
      label: flockGetString(SB, "favs.bar1.link1"),
      callback: function () {
        toggleSidebar("viewBookmarksSidebar", true);
      }
    }];
    message = flockGetString(SB, "favs.bar2.label", [svc.title]);
    
  }else if(svc instanceof Components.interfaces.flockIMediaWebService)
  {
     message = flockGetString(SB, "media.bar2.label", [appName, svc.title]);
    
  }else if(svc instanceof Components.interfaces.flockIBlogWebService)
  {
    message = flockGetString(SB, "blog.bar2.label", [svc.title]);
    link = [{
      label: flockGetString(SB, "blog.bar2.link1"),
      callback: function () {
        flock_blogNewPost();
      }
    }];
    
  }else
  {
    message = this.strbundle.getFormattedString("flock.services.accountsSidebar.configuredAccount", [svc.title]);
  }
  var favicon = "chrome://browser/skin/Info.png";
  try { favicon = document.getElementById("page-proxy-favicon").src; }
  catch (ex) { dump(ex +'\n')};
  
  var notificationBox = this.getNotificationBox();
  notificationBox.removeAllNotifications();
  notificationBox.appendUniqueNotification( message,
    "account-setup-success-" + svc.shortName,
    favicon,
    notificationBox.FLOCK_PRIORITY_MEDIUM, null, link );
}

//! Navigate to account page.
/*!
  Opens the selected accounts home page in a new tab.
*/
accountsSidebar.navigateToAccountPage =
function as_navigateToAccountPage(accountPageUrlLabel) {
  openUILinkIn(accountPageUrlLabel.getAttribute("urlLink"), "tab");
}

//! Log into a service
/*!
 This will just navigate to the service page
*/
accountsSidebar.logIn =
function as_logIn(aAccountCard, aEvent) {

  // Ignore right mouse click
  if (aEvent.button == 2) {
    return;
  }

  var urn = aAccountCard.parentNode.id;
  if (this.acUtils.ensureNoAuthenticatedAccountsBeforeLogin(urn)) {
    var contractID = this.acUtils.getServiceIDForAccountURN(urn);
    var svc = CC[contractID].getService(CI.flockIWebService);
    openUILinkIn(svc.url, "tab");
  }
}

accountsSidebar.logout =
function as_logout(aAccountCard, aEvent) {
  
  // Ignore right mouse click
  if (aEvent.button == 2) {
    return;
  }
  
  this._logger.debug(".logout()");
  var id = aAccountCard.parentNode.id;
  var contractId = this.acUtils.getServiceIDForAccountURN(id);
  var svc = CC[contractId].getService(CI.flockIWebService);
  svc.getAccount(id).logout(null);
  this.acUtils.markAllAccountsAsLoggedOut(contractId);
}

/**
 * Takes a comma-delimited list of sidebar areas to reveal.  Any areas listed
 * in aRevealAreas will be expanded, and any other areas will be collapsed.
 */
accountsSidebar.openAccountArea =
function as_openAccountArea(aRevealAreas) {
  var revealArray = aRevealAreas.split(",");
  for each (var sbArea in this.gAccountsArea) {
    var sbAreaReveal = "false";
    for each (var revealArea in revealArray) {
      if (revealArea == sbArea) {
        sbAreaReveal = "true";
      }
    }
    document.getElementById("flock.services.accountsSidebar.groups." + sbArea)
            .setAttribute("open", sbAreaReveal);
  }
}

accountsSidebar.openBlogSettings =
function as_openBlogSettings(aAccountCard) {
  var id = aAccountCard.parentNode.id;
  var win = window.QueryInterface(CI.nsIInterfaceRequestor)
                  .getInterface(CI.nsIWebNavigation)
                  .QueryInterface(CI.nsIDocShellTreeItem)
                  .rootTreeItem
                  .QueryInterface(CI.nsIInterfaceRequestor)
                  .getInterface(CI.nsIDOMWindow);

  win.openDialog("chrome://flock/content/blog/blogAccountDlg.xul",
                 "editCustomBlog",
                 "modal,centerscreen",
                 { accountId: id });
}
