// vim: ts=2 sw=2 expandtab cindent
//
// BEGIN FLOCK GPL
//
// Copyright Flock Inc. 2005-2008
// http://flock.com
//
// This file may be used under the terms of the
// GNU General Public License Version 2 or later (the "GPL"),
// http://www.gnu.org/licenses/gpl.html
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
// for the specific language governing rights and limitations under the
// License.
//
// END FLOCK GPL

const CC = Components.classes;
const CI = Components.interfaces;
const CR = Components.results;
const CU = Components.utils;

const FLOCK_SNOWMAN_URL = "chrome://flock/skin/people/no_avatar.png";

const DEMO_VIDEO_URL = "http://www.flock.com/tutorials/people/";
const FLOCK_API_URL = "http://developer.flock.com/";

const FLOCK_RICH_CONTENT_FLAVOUR = "application/x-flock-rich-content";
const FLOCK_UNICODE_FLAVOUR = "text/unicode";
const FLOCK_SHELFICON_FLAVOUR = "flock/shelficon";

// constants for peopleSidebarDeck
const DECK_SHOWING_DISCOVERY_PANEL = 0;
const DECK_SHOWING_FRIENDS_LIST = 1;

const LAST_CHANGE_TIMER_INTERVAL = 60 * 1000;
const LUCENE_MAX_RESULTS = 5000;

CU.import("resource:///modules/FlockScheduler.jsm");

var gLoaded = false;
var gFriendListController;
var gCurrentItem = null;
var _coop = CC["@flock.com/singleton;1"]
            .getService(CI.flockISingleton)
            .getSingleton("chrome://flock/content/common/load-faves-coop.js")
            .wrappedJSObject;

var _logger = CC["@flock.com/logger;1"].createInstance(CI.flockILogger);
_logger.init("peopleSidebar");

function loadLibraryFromSpec(aSpec) {
  var loader = CC["@mozilla.org/moz/jssubscript-loader;1"]
               .getService(CI.mozIJSSubScriptLoader);
  loader.loadSubScript(aSpec);
}

loadLibraryFromSpec("chrome://flock/content/common/flocksafe.js");

var _metrics = CC["@flock.com/metrics-service;1"]
               .getService(CI.flockIMetricsService);

// Associative array of accountURN => flockISocialWebServiceAccount objects
var gAccounts = [];

var _wm = CC["@mozilla.org/appshell/window-mediator;1"]
          .getService(CI.nsIWindowMediator);

var _rdfSvc = CC["@mozilla.org/rdf/rdf-service;1"]
              .getService(CI.nsIRDFService);

var _ds = _rdfSvc.GetDataSource("rdf:flock-favorites");

var _acUtils = CC["@flock.com/account-utils;1"]
               .getService(CI.flockIAccountUtils);

var gFilterResults = 0;
var gFilterLastFiltered = 0;
var _sb = CC["@mozilla.org/intl/stringbundle;1"]
          .getService(CI.nsIStringBundleService)
          .createBundle("chrome://flock/locale/people/people.properties");


var gAccountObserver = {
  onAccountAuthenticated: function acctObs_onAccountAuthenticated(aAccountUrn) {
    var coopObj = _coop.get(aAccountUrn);
    var psTabs = document.getElementById("peopleServiceTabs");
    psTabs.setAttribute("collapsed", "false");
    psTabs.buildTabs();
    psTabs.refreshTabData();
    psTabs.selectTabByAccount(coopObj);
    refreshDiscoveryPanel();
    psTabs.toggleRefreshingIndicatorByServiceId(coopObj.serviceId, true);
    gFriendListController._updateAccount(coopObj);
    doLayout();
  },
  onAccountUnauthenticated: function acctObs_onAccountUnauthenticated(aAccountUrn) {
    var coopObj = _coop.get(aAccountUrn);
    this.onAccountNoLongerRefreshing(aAccountUrn);
    var psTabs = document.getElementById("peopleServiceTabs");
    if (psTabs.hasAttribute("lastServiceIdSelected")) {
      // Select the last selected service tab
      psTabs.selectTabByServiceId(psTabs.getAttribute("lastServiceIdSelected"));
    }
    gFriendListController._updateAccount(coopObj);
    refreshDiscoveryPanel();
    doLayout();
  },
  onAccountRefreshing: function acctObs_onAccountRefreshing(aAccountUrn) {
    var coopObj = _coop.get(aAccountUrn);
    var psTabs = document.getElementById("peopleServiceTabs");
    psTabs.toggleRefreshingIndicatorByServiceId(coopObj.serviceId, true);
  },
  onAccountNoLongerRefreshing: function acctObs_onAccountNoLongerRefreshing(aAccountUrn) {
    var coopObj = _coop.get(aAccountUrn);
    var psTabs = document.getElementById("peopleServiceTabs");
    psTabs.toggleRefreshingIndicatorByServiceId(coopObj.serviceId, false);
    psTabs.refreshTabData();
    var filterText = document.getElementById("filterBox").value;
    if (filterText != _sb.GetStringFromName("flock.people.sidebar.filter.hint")) {
      applyFilter(filterText);
    }
  }
};

function FriendListController() {
  this.friendList = document.getElementById("friendList");
  this.order = 1;
  this.sorted = true;
  this.friendIDMap = {};
  this.filterMatches = {};
  this.schedulerTimers = [];
  var inst = this;
  this.peopleObservor = {
    onFriendRemove: function peepObs_onFriendRemove(aFriendUrn) {
      inst._removeFriend(aFriendUrn);
    },
    onFriendUpdate: function peepObs_onFriendUpdate(aFriendUrn) {
      var coopObj = _coop.get(aFriendUrn);
      var timer = CC["@mozilla.org/timer;1"].createInstance(CI.nsITimer);
      var cb = {
        notify: function cb_notify() {
          inst._updateFriend(coopObj);
        }
      }
      timer.initWithCallback(cb, 0, CI.nsITimer.TYPE_ONE_SHOT);
    },
    onInsertFriend: function peepObs_onInsertFriend(aFriendUrn) {
      var coopObj = _coop.get(aFriendUrn);
      if (coopObj.isInstanceOf(_coop.Identity)) {
        var timer = CC["@mozilla.org/timer;1"].createInstance(CI.nsITimer);
        var cb = {
          notify: function FriendListController_assert_notify() {
            inst._insertFriend(coopObj, coopObj.getParent().account);
          }
        }
        timer.initWithCallback(cb, 0, CI.nsITimer.TYPE_ONE_SHOT);
      }
    }
  };
  this.lastChangeTimer = CC["@mozilla.org/timer;1"].createInstance(CI.nsITimer);
  this.lastChangeTimerListener = {
    notify: function FriendListController_notify(aTimer) {
      _logger.debug("FriendListController: lastChangeTimerListener");
      function updateTimes(aShouldYield) {
        var nodeList = inst.friendList.childNodes;
        for (var c = 0; c < nodeList.length; c++) {
          nodeList.item(c).updateTime();
          if (aShouldYield()) {
            yield;
          }
        }
      }
      FlockScheduler.schedule(inst.schedulerTimers, 0.05, 10, updateTimes);
    }
  }
}
FriendListController.prototype.init =
FriendListController.prototype.init =
function FriendListController_init() {
  var accounts = _coop.accounts_root.children.enumerate();
  while (accounts.hasMoreElements()) {
    var account = accounts.getNext();
    var service = CC[account.serviceId].getService(CI.flockIWebService);
    if (service instanceof CI.flockISocialWebService) {
      this._updateAccount(account);
    }
  }

  this.lastChangeTimer.initWithCallback(this.lastChangeTimerListener,
                                        LAST_CHANGE_TIMER_INTERVAL,
                                        CI.nsITimer.TYPE_REPEATING_SLACK);
  var p = CC["@flock.com/people-service;1"].getService(CI.flockIPeopleService);
  p.addObserver(this.peopleObservor);
}

FriendListController.prototype.uninit =
function FriendListController_uninit() {
  this.lastChangeTimer.cancel();
  this.lastChangeTimer = null;
  for (var c = this.schedulerTimers.length - 1; c >= 0; c--) {
    FlockScheduler.cancel(this.schedulerTimers, c);
  }
  var p = CC["@flock.com/people-service;1"].getService(CI.flockIPeopleService);
  p.removeObserver(this.peopleObservor);
}
FriendListController.prototype._compare =
function FriendListController__compare(friendCard, refCard) {
  if (this.order == 1) {
    var friendLastUpdate = friendCard.getAttribute("lastupdate");
    var refLastUpdate = refCard.getAttribute("lastupdate");
    var friendLastUpdateInt = Number(friendLastUpdate);
    var refLastUpdateInt = Number(refLastUpdate);
    var friendName = friendCard.getAttribute("displayName").toLowerCase();
    var refName = refCard.getAttribute("displayName").toLowerCase();
    if (friendLastUpdateInt > refLastUpdateInt ||
        (friendLastUpdateInt == refLastUpdateInt && friendName <= refName))
    {
      return true;
    } else {
      return false;
    }
  } else if (this.order == 2) {
    var friendName = friendCard.getAttribute("displayName").toLowerCase();
    var refName = refCard.getAttribute("displayName").toLowerCase();
    if (friendName <= refName) {
      return true;
    } else {
      return false;
    }
  } else {
    _logger.error("FriendListController: _compare: invalid order: " + this.order);
    return false;
  }
}

FriendListController.prototype._findPosition =
function FriendListController__findPosition(friendCard) {
  var position = 0;
  for (var c = 0; c < this.friendList.childNodes.length; c++) {
    var refCard = this.friendList.childNodes.item(c);
    if (this._compare(friendCard, refCard)) {
      break;
    }
    position = c + 1;
  }
  return position;
}

FriendListController.prototype._findPositionSorted =
function FriendListController__findPositionSorted(friendCard) {
  var low = 0;
  var high = this.friendList.childNodes.length - 1;
  var mid, refCard;
  while (low < high) {
    mid = Math.floor((low + high) / 2);
    refCard = this.friendList.childNodes.item(mid);
    if (this._compare(friendCard, refCard)) {
      high = mid;
    } else {
      low = mid + 1;
    }
  }
  if (low == high) {
    refCard = this.friendList.childNodes.item(low);
    if (!this._compare(friendCard, refCard)) {
      low++;
    }
  }
  return low;
}

FriendListController.prototype._insertFriend =
function FriendListController__insertFriend(friend, account) {
  _logger.debug("FriendListController: _insertFriend");
  if (!account) {
    return;
  }

  // check for existing friend card element
  var friendCard = this.friendIDMap[friend.id()];
  if (friendCard) {
    return;
  }

  friendCard = document.createElement("friendcard");

  friendCard.setAttribute("accountURN", account.id());
  friendCard.setAttribute("accountId", friend.accountId);
  friendCard.setAttribute("displayName", friend.name);
  friendCard.setAttribute("avatar", friend.avatar);
  friendCard.setAttribute("avatarcommand", "");
  friendCard.setAttribute("lastupdate", friend.lastUpdate);
  friendCard.setAttribute("lastupdatetype", friend.lastUpdateType);
  friendCard.setAttribute("quote", friend.statusMessage);
  friendCard.setAttribute("serviceId", friend.serviceId);
  friendCard.setAttribute("unseenMedia", friend.unseenMedia);
  friendCard.setAttribute("uri", friend.id());
  friendCard.setAttribute("id", friend.id());

  friendCard.setAttribute("ondragdrop",
                          "this.removeAttribute('dropzoneengaged');"
                          + "Components.classes['@flock.com/metrics-service;1']"
                          + ".getService(Components.interfaces.flockIMetricsService)"
                          + ".report('PeopleSidebar-FriendCardDnD', 'General');"
                          + "nsDragAndDrop.drop(event, personcardDNDObserver, true);"
                          + "event.stopPropagation();");
  friendCard.setAttribute("ondragover",
                          "this.setAttribute('dropzoneengaged', 'true');"
                          + "nsDragAndDrop.dragOver(event, personcardDNDObserver);"
                          + "event.stopPropagation();");
  friendCard.setAttribute("ondragenter",
                          "this.setAttribute('dropzoneengaged', 'true');"
                          + "nsDragAndDrop.dragEnter(event, personcardDNDObserver);"
                          + "event.stopPropagation();");
  friendCard.setAttribute("ondragexit",
                          "this.removeAttribute('dropzoneengaged');"
                          + "nsDragAndDrop.dragExit(event, personcardDNDObserver);"
                          + "event.stopPropagation();");

  var position;
  if (this.sorted) {
    position = this._findPositionSorted(friendCard);
  } else {
    position = this._findPosition(friendCard);
  }
  if (position == this.friendList.childNodes.length) {
    this.friendList.appendChild(friendCard);
  } else {
    var refCard = this.friendList.childNodes.item(position);
    this.friendList.insertBefore(friendCard, refCard);
  }

  friendCard.displayName = friend.name;
  this.friendIDMap[friend.id()] = friendCard;
}

FriendListController.prototype._removeFriend =
function FriendListController__removeFriend(friendURI) {
  _logger.debug("FriendListController: _removeFriend");
  var friendCard = this.friendIDMap[friendURI];
  if (friendCard) {
    this.friendList.removeChild(friendCard);
    delete this.friendIDMap[friendURI];
  }
}

FriendListController.prototype._updateFriend =
function FriendListController__updateFriend(identity) {
  _logger.debug("FriendListController: _updateFriend");
  var account = identity.getParent().account;
  this._removeFriend(identity.id());
  this._insertFriend(identity, account);
}

FriendListController.prototype._updateAccount =
function FriendListController__updateAccount(account) {
  _logger.debug("FriendListController: _updateAccount");
  var inst = this;
  function updateAccountContent(aShouldYield) {
    if (account.isAuthenticated) {
      var friends = account.friendsList.children.enumerate();
      while (friends.hasMoreElements()) {
        var friend = friends.getNext();
        inst._insertFriend(friend, account);
        if (aShouldYield()) {
          yield;
        }
      }
    } else {
      var friends = account.friendsList.children.enumerate();
      while (friends.hasMoreElements()) {
        var friend = friends.getNext();
        inst._removeFriend(friend.id());
        if (aShouldYield()) {
          yield;
        }
      }
    }
  }
  FlockScheduler.schedule(this.schedulerTimers, 0.1, 50, updateAccountContent);
}

FriendListController.prototype.sort =
function FriendListController_sort(order) {
  _logger.debug("FriendListController: sort");
  if (this.order == order) {
    return;
  }
  this.order = order;
  this.sorted = false;
  var inst = this;
  function sortContent(aShouldYield) {
    for (var c = 0; c < inst.friendList.childNodes.length; c++) {
      var friendCard = inst.friendList.childNodes.item(c);
      var newPosition = inst._findPosition(friendCard);
      if (newPosition != c) {
        if (newPosition == inst.friendList.childNodes.length) {
          inst.friendList.appendChild(friendCard);
        } else {
          var refCard = inst.friendList.childNodes.item(newPosition);
          inst.friendList.insertBefore(friendCard, refCard);
        }
      }
      if (aShouldYield()) {
        yield;
      }
    }
    inst.sorted = true;
  }
  FlockScheduler.schedule(this.schedulerTimers, 0.25, 80, sortContent);
}

FriendListController.prototype.setFilter =
function FriendListController_setFilter(aLuceneResults) {
  _logger.debug("FriendListController: setFilter");
  var newFilterMatches = {};
  while (aLuceneResults.hasMoreElements()) {
    var result = aLuceneResults.getNext().QueryInterface(CI.flockILuceneResult);
    var friendCard = this.friendIDMap[result.URI];
    if (friendCard) {
      friendCard.setAttribute("class", "match");
      newFilterMatches[result.URI] = true;
    }
  }
  for (var id in this.filterMatches) {
    if (!newFilterMatches[id]) {
      var friendCard = this.friendIDMap[id];
      friendCard.setAttribute("class", "");
    }
  }
  this.filterMatches = newFilterMatches;
  this.friendList.setAttribute("class", "filtered");
}

FriendListController.prototype.clearFilter =
function FriendListController_clearFilter() {
  _logger.debug("FriendListController: clearFilter");
  this.friendList.setAttribute("class", "");
}


function performActionSpec(aActionSpec, aAccountURN, aFriendURN, aTransferable) {
  _logger.debug("performActionSpec");
  _logger.debug("aActionSpec = " + aActionSpec);
  _logger.debug("aAccountURN = " + aAccountURN);
  _logger.debug("aFriendURN = " + aFriendURN);
  // The action spec has four possible values:
  //  "newtab:<url>" - open a new tab with the specified URL
  //  "newwindow:<url>" - open a new window with the specified URL
  //  "mediaquery:<query>" - open the media bar with the specified stream
  //  "method:<interface>:<methodname>" - QI the service to the specified
  //      interface and then call the specified method. The callback method
  //      will always have (aFriendURN, aTransferable) passed in as parameters.

  // Parse type and action out of the spec
  var colonIndex = aActionSpec.indexOf(":");
  if (colonIndex == -1) {
    _logger.debug("Invalid action spec: " + aActionSpec);
    return;
  }
  var specType = aActionSpec.substring(0, colonIndex);
  var specAction = aActionSpec.substring(colonIndex + 1);

  // Handle each scenario
  switch (specType) {
    case "newtab":
      loadURIIn(specAction, "tab");
      break;

    case "newwindow":
      loadURIIn(specAction, "window");
      break;

    case "mediaquery":
      // TODO
      break;

    case "method":
      // Get interface and methodname
      colonIndex = specAction.indexOf(":");
      if (colonIndex == -1) {
        _logger.debug("Invalid method action: " + specAction);
        return;
      }
      var interface = specAction.substring(0, colonIndex);
      var methodname = specAction.substring(colonIndex + 1);
      // Determine which service this friend is from
      var serviceId;
      var accountURN;
      if (aAccountURN) {
        var account = _coop.get(aAccountURN);
        accountURN = aAccountURN;
        serviceId = account.serviceId;
      } else {
        var identity = _coop.get(aFriendURN);
        accountURN = identity.getParent().account.id();
        serviceId = identity.serviceId;
      }

      // Get the service account
      var srvAcct = CC[serviceId]
                    .getService(CI.flockIWebService)
                    .getAccount(accountURN)
                    .QueryInterface(CI[interface])
      // Call its method
      if (aAccountURN) {
        srvAcct[methodname](aTransferable);
      } else {
        srvAcct[methodname](aFriendURN, aTransferable);
      }
      break;

      case "mediabar":
        colonIndex = specAction.indexOf(":");
        if (colonIndex == -1) {
          _logger.debug("Invalid method action: " + specAction);
          return;
        }
        var queryType = specAction.substring(0, colonIndex);
        var identity = _coop.get(aFriendURN);
        var srvName = identity.getParent().account.service.name;
        if (!queryType) {
          var query = new queryHelper();
          query.user = identity.accountId;
          query.username = identity.name;
          query = query.stringVal;
        } else {
          query = queryType + ":" + identity.accountId;
        }
        _openMediaBar(srvName, query, identity.name);
        break;

    default:
      _logger.debug("Unknown specType: " + specType);
      return;
  }
}

// Assemble the tag line tooltip for a mecard or friendcard
function assembleTooltip(aNode) {
  _logger.debug("assembleTooltip");
  var tooltipDisplayNameBroadcaster = document.getElementById("tooltipDisplayNameBroadcaster");
  tooltipDisplayNameBroadcaster.setAttribute("value", aNode.displayName);

  var tooltipTaglineBroadcaster = document.getElementById("tooltipTaglineBroadcaster");
  tooltipTaglineBroadcaster.setAttribute("value", aNode.tagline);

  // Set the height of the tooltip popup (for some reason the tooltip popup
  // does not expand to fit the contents without doing this)
  var tooltip = document.getElementById("taglineTooltip");
  var container = document.getElementById("ttipOuterContainer");
  tooltip.height = container.boxObject.height + 10;
}

function getCurrentTabURL() {
  _logger.debug("getCurrentTabURL()")
  var win = _wm.getMostRecentWindow("navigator:browser");
  var result = null;
  if (win) {
    var gBrowser = win.getBrowser();
    result = gBrowser.currentURI.spec;
  }
  _logger.debug("result = " + result);
  return result;
}

function toggleObservers(aToggleOn) {
  _logger.debug("toggleObservers(" + aToggleOn + ")");
  var accountService = CC["@flock.com/account-service;1"]
                       .getService(CI.flockIAccountService);
  if (aToggleOn) {
    accountService.addObserver(gAccountObserver);
  } else {
    accountService.removeObserver(gAccountObserver);
  }
}

// Auto-remember account when we click a service for discovery
function rememberAccount(aSvcURN) {
  _logger.debug("rememberAccount: " + aSvcURN);
  var wsMgr = CC["@flock.com/webservice-manager;1"]
              .getService(CI.flockIWebServiceManager);
  wsMgr.autoKeepAccountForService(aSvcURN);
}

// Load a url in a new tab or window
function loadURIIn(aUri, aWhere) {
  _logger.debug("loadURIIn('" + aUri + "', '"+ aWhere +"')");
  var win = _wm.getMostRecentWindow("navigator:browser");
  if (win) {
    win.openUILinkIn(aUri, aWhere);
  }
}

function onKeyup(aEvent) {
  if (aEvent.keyCode == KeyEvent.DOM_VK_RETURN) {
    aEvent.originalTarget.loadProfile();
  }
}

function loadURI(aUri, aEvent) {
  _logger.debug("loadURI(" + aUri + ")");
  if (aEvent.button != 2) {
    // Left or Middle clicks ok
    var openIn = "current";

    if (aEvent.button == 1) {
      // Middle click
      openIn = "tab";
    }
//@line 619 "/home/build/tinderbuild_blue/src/flock/mozilla/flock/base/people/content/peopleSidebar.js"
    if (aEvent.ctrlKey) {
      // Ctrl Key held down
      openIn = "tab";
    }
//@line 624 "/home/build/tinderbuild_blue/src/flock/mozilla/flock/base/people/content/peopleSidebar.js"
    if (aEvent.shiftKey) {
      // Shift key held down
      openIn = "window";
    }
    var win = _wm.getMostRecentWindow("navigator:browser");
    if (win &&
        (openIn == "current") &&
        (win.gBrowser.currentURI.spec == "about:myworld")) {
      // Never load over MyWorld.  Load in new tab
      openIn = "tab";
    }

    win.openUILinkIn(aUri, openIn);
  }
}

function logout(aNode, aEvent) {
  _logger.debug("logout()");
  var serviceId = aNode.getAttribute("serviceId");
  CC[serviceId].getService(CI.flockIWebService).logout();
  _acUtils.markAllAccountsAsLoggedOut(serviceId);
}

// Display a media item in the media bar
function openMediaBar(aNode, aEvent) {
  _logger.debug("openMediaBar()");
  var name = aNode.getAttribute("displayName");
  _logger.debug("name: " + name);
  var accountId = aNode.getAttribute("accountId");
  _logger.debug("accountId: " + accountId);

  var parentID = aNode.getAttribute("id");
  var identity = _coop.get(parentID);
  var accountURN = identity.getParent().account.id();
  var serviceId = identity.serviceId;

  var svc = CC[identity.serviceId]
            .getService(CI.flockISocialWebService);
  svc.markAllMediaSeen(aNode.id);

  var query = new queryHelper();
  query.user = accountId;
  query.username = name;

  _openMediaBar(svc.shortName, query.stringVal, name);

  // Pop message bar
  var favItem = _coop.MediaQuery.find({query: query.stringVal, isPollable: true});
  if (favItem.length == 0) {
    var msg = flockGetString("favorites/favorites",
                             "flock.favs.media.discovery");
    var nBox = getNotificationBox();
    nBox.appendUniqueNotification(msg,
                                  "media-discovery-info",
                                  "chrome://browser/skin/flock/favorites/media-icon.png",
                                  nBox.FLOCK_PRIORITY_LOW,
                                  null,
                                  null);
  }
}

function getNotificationBox() {
  var wm = CC["@mozilla.org/appshell/window-mediator;1"]
           .getService(CI.nsIWindowMediator);
  var win = wm.getMostRecentWindow("navigator:browser");
  if (win) {
    var gBrowser = win.getBrowser();
    var notificationBox = gBrowser.getNotificationBox();

    return notificationBox;
  }

  return null;
}

function _openMediaBar(aShortName, aStringValue, aName)
{
  var win = _wm.getMostRecentWindow("navigator:browser");
  if (win) {
    win.flock_topbarSelectById("flock_MediabarTopbarBroadcaster", true);
    win.mediabar.loadQuery(aShortName, aStringValue, aName);
  }
}

function applyFilter(aFilterText) {
  _logger.debug("applyFilter");
  var filterBox = document.getElementById("filterBox");
  if (aFilterText) {
    var searchService = CC["@flock.com/lucene/flockLucene;1"]
                        .getService(CI.flockILucene);
    var inst = this;
    var searchListener = {
      onSearchComplete: function(aNumResults, aResults) {
        _logger.debug("onSearchComplete: " + aNumResults);
        gFilterResults = aNumResults;
        gFilterLastFiltered = Date.now();
        gFriendListController.setFilter(aResults);
      }
    }
    _logger.debug("applyFilter: executing query: " + normalizeQuery(aFilterText));
    searchService.search(normalizeQuery(aFilterText),
                         "identity",
                         LUCENE_MAX_RESULTS,
                         searchListener);
  } else {
    gFriendListController.clearFilter();
  }
}

function serviceTabClicked() {
  _logger.debug("serviceTabClicked");
  if (!gLoaded) {
    return;
  }

  // Keep track of the tab selected so that we can select if the user opens
  // and closes the sidebar
  var pst = document.getElementById("peopleServiceTabs");
  pst.setAttribute("lastServiceIdSelected", pst.selectedServiceId);
  
  // Report which tab was selected to the metrics service
  var tabName = "all";
  if (pst.selectedServiceId != null) {
    tabName = CC[pst.selectedServiceId]
              .getService(CI.flockISocialWebService).shortName;
  }
  _metrics.report("PeopleSidebar-ServiceClick", tabName);

  updateFriendListFilter();
  doLayout();
}

function updateFriendListFilter() {
  // Set the currentServiceId attribute on the friendlist as this will be
  // referenced in the css so that only friends appropriate to the selected
  // tab are shown
  var pst = document.getElementById("peopleServiceTabs");
  var friendsList = document.getElementById("friendList");
  if (pst.selectedServiceId) {
    friendsList.setAttribute("currentServiceId", pst.selectedServiceId);
  } else {
    friendsList.removeAttribute("currentServiceId");
  }
}

function onLoad() {
  _logger.debug("onLoad");
  toggleObservers(true);

  gFriendListController = new FriendListController();
  gFriendListController.init();

  _metrics.report("PeopleSidebar-Opened", null);
  // get the persisting sortmode
  var sortmode = document.getElementById("peopleSidebar")
                         .getAttribute("sortmode");
  document.getElementById("sort-menu")
          .selectedItem = document.getElementById(sortmode);

  var psTabs = document.getElementById("peopleServiceTabs");

  // Only build tabs if there are accounts to be shown
  var accts = _coop.accounts_root.children.enumerate();
  var acctsExist = false;
  while (accts.hasMoreElements()) {
    var coopAcct = accts.getNext();
    // Get a reference to the service
    var service = CC[coopAcct.serviceId].getService(CI.flockIWebService);
    if (service instanceof CI.flockISocialWebService) {
      acctsExist = true;
      break;
    }
  }

  if (acctsExist) {
    // Accounts exist, build the tabs
    psTabs.buildTabs();
    psTabs.refreshTabData();
  } else {
    // Make sure we hide the All tab
    psTabs.setAttribute("collapsed", "true");
  }

  var psbBroadcaster = top.document.getElementById("flockPeopleSidebarBroadcaster");
  if (psbBroadcaster.hasAttribute("newAccountId")) {
    // Select the service tab for a newly added account
    psTabs.selectTabByAccountId(psbBroadcaster.getAttribute("newAccountId"));
    psbBroadcaster.removeAttribute("newAccountId");

    // in the case where a new account has been configured and the sidebar
    // is closed,  we need to toggle the throbber on manually because the
    // isRefreshing=true happens before the rdf observer in the sidebar is
    // created
    psTabs.toggleRefreshingIndicatorByServiceId(psTabs.selectedAccount
                                                      .serviceId,
                                                true);
  } else if (psTabs.hasAttribute("lastServiceIdSelected")) {
    // Select the last selected service tab
    psTabs.selectTabByServiceId(psTabs.getAttribute("lastServiceIdSelected"));
  }
  updateFriendListFilter();

  refreshDiscoveryPanel();
  doLayout();
  //setTimeout("gFriendListController.init();", 100);
  setTimeout("refreshSelectedAccount();", 2000);
  gLoaded = true;
}

function refreshSelectedAccount() {
  var psTabs = document.getElementById("peopleServiceTabs");
  if (psTabs.selectedAccount) {
    CC["@flock.com/poller-service;1"]
    .getService(CI.flockIPollerService)
    .forceRefresh(psTabs.selectedAccount.id());
  }
}

// Set the visibility of the cancel filter button
function layoutCancelFilterButton() {
  var filter = document.getElementById("filterBox");
  // If the user has entered some filter text then display the cancel filter button
  if (filter.value && (filter.value != _sb.GetStringFromName("flock.people.sidebar.filter.hint"))) {
    document.getElementById("cancelFilterContainer").setAttribute("hidden", "false");
  } else {
    document.getElementById("cancelFilterContainer").setAttribute("hidden", "true");
  }
}

// Configure the friends list based on the service and
// account being displayed, in addtion to laying out the
// UI
function doLayout() {
  _logger.debug("doLayout start");
  var serviceTabs = document.getElementById("peopleServiceTabs");
  var isloggedInAccounts = isAnAccountLoggedIn();
  var isAllCardSelected = serviceTabs.isAllCardSelected;
  var showFriendsList = false;
  if (isAllCardSelected) {
    if (isloggedInAccounts) {
      // Show the friends list as the All card is selected and at least one
      // account is logged in
      showFriendsList = true;
    }
  } else {
    _logger.debug("Selected serviceId: " + serviceTabs.selectedServiceId);
    var displayAccountURN = _acUtils.
                            getFirstAuthenticatedAccountForService(serviceTabs
                                                                  .selectedServiceId);
    if (displayAccountURN) {
      // Display the friends list as a service tab has been selected and the
      // account is logged in
      showFriendsList = true;
    }
  }

  if (showFriendsList) {
    sortFriendsList();
    // Show the friends list
    document.getElementById("peopleSidebarDeck")
            .selectedIndex = DECK_SHOWING_FRIENDS_LIST;
  } else {
    _logger.debug("Showing the discovery panel");
    document.getElementById("peopleSidebarDeck")
            .selectedIndex = DECK_SHOWING_DISCOVERY_PANEL;
  }

  _logger.debug("doLayout end");
}

function getFormattedTime(aInput) {

  var dateFormatter = CC["@flock.com/date-formatter;1"]
                      .getService(CI.flockIDateFormatter);
  var now = new Date();
  var seconds = Math.round(now.getTime() / 1000) - aInput;
  var minutes = Math.round(seconds / 60);
  var hours = Math.round(seconds / (60 * 60));
  var days = Math.round(seconds / (60 * 60 * 24));

  if (seconds < 60) {
    return _sb.GetStringFromName("flock.people.mecard.lastUpdatePretty.aMomentAgo");
  } else if (minutes < 60) {
    if (minutes > 1) {
      return minutes + 
             " " + 
             _sb.GetStringFromName("flock.people.mecard.lastUpdatePretty.minutesAgo");
    } else {
      return _sb.GetStringFromName("flock.people.mecard.lastUpdatePretty.oneMinuteAgo");
    }
  } else if (hours < 24) {
    if (hours > 1) {
      return hours + 
             " " + 
             _sb.GetStringFromName("flock.people.mecard.lastUpdatePretty.hoursAgo");
    } else {
      return _sb.GetStringFromName("flock.people.mecard.lastUpdatePretty.oneHourAgo");
    }
  } else if (days < 7) {
    if (days > 1) {
      return days + 
             " " + 
             _sb.GetStringFromName("flock.people.mecard.lastUpdatePretty.daysAgo");
    } else {
      return _sb.GetStringFromName("flock.people.mecard.lastUpdatePretty.oneDayAgo");
    }
  } else {
    return dateFormatter.getShortDateTimeString(aInput * 1000);
  }
}

function getFormattedUpdateType(aLastUpdateType, aStatus) {
  switch (aLastUpdateType) {
    case "profile": {
      return _sb.GetStringFromName("flock.people.mecard.lastUpdateTypePretty.profile");
      break;
    }
    case "status": {
      if (aStatus == "") {
        return _sb.GetStringFromName("flock.people.mecard.lastUpdateTypePretty.statusCleared");
      } else {
        return _sb.GetStringFromName("flock.people.mecard.lastUpdateTypePretty.statusUpdated");
      }
      break;
    }
    case "media": {
      return _sb.GetStringFromName("flock.people.mecard.lastUpdateTypePretty.mediaUpdated");
      break;
    }
  }
}

function refreshDiscoveryPanel() {
  _logger.debug("refreshDiscoveryPanel");
  // refresh available services for discovery
  var catMgr = CC["@mozilla.org/categorymanager;1"].getService(CI.nsICategoryManager);
  var e = catMgr.enumerateCategory("flockWebService");
  var serviceBox = document.getElementById("available-services-content");
  var isRemainingUnconfiguredServices = false;
  var svcChild = serviceBox.firstChild;
  while (svcChild) {
    serviceBox.removeChild(svcChild);
    svcChild = serviceBox.firstChild;
  }

  while (e.hasMoreElements()) {
    var entry = e.getNext().QueryInterface(CI.nsISupportsCString);
    if (!entry) continue;

    var contractID = catMgr.getCategoryEntry("flockWebService", entry.data);
    var svc = CC[contractID].getService(CI.flockIWebService);

    if (svc instanceof CI.flockISocialWebService) {
      _logger.debug("refreshDiscoveryPanel(): Found a service - " 
                    + svc.shortName);

      // Skip services which already have accounts configured.
      var accountsEnum = svc.getAccounts();
      if (accountsEnum.hasMoreElements()) {
        continue;
      }

      isRemainingUnconfiguredServices = true;
      var favicon = document.createElement("image");
      favicon.setAttribute("src", svc.icon);

      var label = document.createElement("label");
      label.setAttribute("id", "discovery-service-link");
      label.setAttribute("value", svc.title
                                  + " - "
                                  + _sb.GetStringFromName("flock.people.sidebar.services.clickMessage"));
      label.setAttribute("onclick",
                         "Components.classes['@flock.com/metrics-service;1']"
                         + ".getService(Components.interfaces.flockIMetricsService)"
                         + ".report('PeopleSidebar-DiscoverService', '" + svc.shortName + "');"
                         + "rememberAccount('" + svc.urn + "');"
                         + "loadURI('http://redir.flock.com/redir?component=browser&action=people_click&partner="
                         + svc.shortName + "', event)");
      label.setAttribute("flex", "1");
      label.setAttribute("crop", "end");

      var hbox = document.createElement("hbox");
      hbox.setAttribute("id", "discovery-service-entry");
      hbox.appendChild(favicon);
      hbox.appendChild(label);
      serviceBox.appendChild(hbox);
    }
  }

  // Suppress the available services content box when the user has a configured
  // account for all services supported in the people sidebar (i.e. flockISocialWebService)
  document.getElementById("content-box-middle")
          .setAttribute("hidden", 
                        (isRemainingUnconfiguredServices ? "false" : "true"));
}

function sortFriendsList() {
  var friendsList = document.getElementById("friendList");
  var selectedSort = document.getElementById("sort-menu").selectedItem;
  if (selectedSort) {
    _logger.debug("set sort: " + selectedSort.value);
    // Remember the sort mode for later
    document.getElementById("peopleSidebar")
            .setAttribute("sortmode", selectedSort.id);
    gFriendListController.sort(selectedSort.value);

    friendsList.selectedIndex = 0;

    _metrics.report("PeopleSidebar-FriendsSorted", 
                    [{sort: selectedSort.id}]);
  }
}

function isAnAccountLoggedIn() {
  _logger.debug("isAnAccountLoggedIn()");
  var accts = _coop.accounts_root.children.enumerate();
  while (accts.hasMoreElements()) {
    var coopAcct = accts.getNext();
    var account = CC[coopAcct.serviceId].getService(CI.flockIWebService)
                                        .getAccount(coopAcct.id());
    if ((account instanceof CI.flockISocialWebServiceAccount)
        && coopAcct.isAuthenticated)
    {
      _logger.debug("Found authenticated account");
      return true;
    }
  }
  _logger.debug("No authenticated account");
  return false;
}

function sortFriends() {
  _logger.debug("sortFriends()");
  var selection = document.getElementById("sort-menu").selectedItem;
  var friendListContainer = document.getElementById("friendList");
  if (selection) {
    _logger.debug("sorting by " + selection.value);
    document.getElementById("peopleSidebar")
            .setAttribute("sortmode", selection.id);
    friendListContainer.setAttribute("sortResource", selection.value);

    var sortDirection = selection.getAttribute("sortDirection");
    friendListContainer.setAttribute("sortDirection", sortDirection);

    friendListContainer.builder.rebuild();
    friendListContainer.selectedIndex = 0;
  }
}

function unLoad() {
  _logger.debug("unLoad");
  if (gFriendListController) {
    toggleObservers(false);
    gFriendListController.uninit();
    _metrics.report("PeopleSidebar-Close", null);
    gFriendListController = null;
  }
}

function normalizeQuery(aInput) {
  var output;
  var in_quotes = false;

  // strip leading spaces
  while (aInput.length && aInput[0] == " ") {
    // chop off the first character
    aInput = aInput.substr(1);
  }

  // Convert to lowercase for the wildcard search
  aInput = aInput.toLowerCase();

  // add leading plus
  output = "+";
  while (aInput.length) {
    if (aInput[0] == "\"") {
      if (in_quotes) {
        in_quotes = false;
      } else {
        in_quotes = true;
        // FIXME: remove spaces from the start of quoted strings
      }
    }

    // escape special chars
    if ("&|!(){}[]^~*?:\\".indexOf(aInput[0]) >= 0) {
      output = output + "\\";
    }

    output = output + aInput[0];

    if (!in_quotes && aInput[0] == " " && aInput.length > 1) {
        // put a plus before every word
        output = output + "+";
    }

    aInput = aInput.substr(1);
  }

  // remove trailing spaces
  var removed_spaces = false;
  while (output.length > 0 && output[output.length - 1] == " ") {
    output = output.substr(0, output.length - 1);
    removed_spaces = true;
  }

  // if we're in a quoted string we should close that
  if (in_quotes) {
    output = output + "\"";
  }

  // if theres an incomplete word
  if ((!removed_spaces) && (output[output.length - 1] != "\"")) {
    // search for partial matches
    output = output + "*";
  }

  return output;
}

function filterOnBlur() {
  var filter = document.getElementById("filterBox");
  if (filter.value == "") {
    filter.setAttribute("empty", "true");
    filter.value = filter.getAttribute("hint");
  } else {
    if ((Date.now() - gFilterLastFiltered) > 300) {
      var filterSize = filter.value.length;
      _metrics.report("PeopleSidebar-FriendsFiltered", 
                      [{results: gFilterResults, size: filterSize}]);
    }
  }
}


function filterOnFocus() {
  var filter = document.getElementById("filterBox");
  if (filter.getAttribute("empty") == "true") {
    filter.removeAttribute("empty");
    filter.value = "";
  }
}

// localizeElement is a helper function that takes a xul element and sets the
// value of a specified attribute to a localized resource string. If aServiceId
// is provided then the function will pull the string from the string bundle
// for the service otherwise it used the default string bundle for the people
// sidebar (people.properties)
// aElement: target element
// aAttributeName: attribute that will be set
// aStringName: resource string name
// aServiceId: service id used for a service specific resource
function localizeElement(aElement,
                         aAttributeName,
                         aStringName,
                         aServiceId)
{
  var bundle;
  if (aServiceId) {
    // Use service bundle
    bundle = CC[aServiceId]
            .getService(CI.flockIWebService)
            .getStringBundle(); 
  } else {
    // Use the default bundle
    bundle = _sb;
  }

  try {
    // Get the service specific string if there is one
    var string = bundle.GetStringFromName(aStringName);
    aElement.setAttribute(aAttributeName, string);
  } catch (e) {
    _logger.error("Exception in call to localizeElement");
    // Do nothing
  }
}

// This is the DND observer for the Post Link button
var mecardPostLinkDNDObserver = {
  canDrop: function mc_canDrop(aEvent, aDragSession) {
    // Cannot drop if dragged from outside browser or text
    var node = aDragSession.sourceNode;
    if (!node) {
      return false;
    }
    var text_node = 3;
    var linkDragged = true;
    if (node.nodeType == text_node) {
      var top = _wm.getMostRecentWindow("navigator:browser");
      var selection = top.getSelection();
      if (selection.toString().length > 0) {
        linkDragged = false;
      }
    }
    return linkDragged;
  },

  canHandleMultipleItems: false,

  onDragOver: function mc_onDragOver(aEvent, aFlavor, aSession) {
    _logger.debug("mecardPostLinkDNDObserver.onDragOver()");
  },

  onDragEnter: function mc_onDragEnter(aEvent, aSession) {
    _logger.debug("mecardPostLinkDNDObserver.onDragEnter()");
  },

  onDragExit: function mc_onDragExit(aEvent, aSession) {
    _logger.debug("mecardPostLinkDNDObserver.onDragExit()");
  },

  onDrop: function mc_onDrop(aEvent, aDropData, aSession) {
    _logger.debug("mecardPostLinkDNDObserver.onDrop()");
    aEvent.stopPropagation();

    // aEvent.target is the "Post Link" button
    var meCard = document.getBindingParent(aEvent.target);

    // Get account and service info
    var serviceId = meCard.serviceId;
    var accountURN = meCard.c_account.id();

    // Get transferable from session
    var transferable = CC["@flock.com/rich-dnd-service;1"]
                       .getService(CI.flockIRichDNDService)
                       .getRichSelection(aSession,
                                         aDropData.flavour.contentType,
                                         aDropData.data);

    // Don't continue if we don't have any data
    if (!transferable) {
      return;
    }

    // Get service-appropriate post link action spec and perform it
    var postLinkAction = CC[serviceId]
                         .getService(CI.flockIWebService)
                         .getAccount(accountURN)
                         .QueryInterface(CI.flockISocialWebServiceAccount)
                         .getPostLinkAction(transferable);

    performActionSpec(postLinkAction, accountURN, null, transferable);
  },

  getSupportedFlavours: function mc_getSupportedFlavours() {
    _logger.debug("mecardPostLinkDNDObserver.getSupportedFlavours()");
    var flavors = new FlavourSet();
    flavors.appendFlavour(FLOCK_RICH_CONTENT_FLAVOUR);
    flavors.appendFlavour("text/x-moz-url");
    flavors.appendFlavour("text/unicode");
    flavors.appendFlavour("text/html");
    return flavors;
  }
}

var personcardDNDObserver = {
  canDrop: function pc_canDrop(aEvent, aDragSession) {
    // Cannot drop if dragged from outside browser
    return (aDragSession.sourceNode != null);
  },

  canHandleMultipleItems: false,

  onDragOver: function pc_onDragOver(aEvent, aFlavor, aSession) {
    _logger.debug("personcardDNDObserver.onDragOver()");
  },

  onDragEnter: function pc_onDragEnter(aEvent, aSession) {
    _logger.debug("personcardDNDObserver.onDragEnter()");
  },

  onDragExit: function pc_onDragExit(aEvent, aSession) {
    _logger.debug("personcardDNDObserver.onDragExit()");
  },

  onDrop: function pc_onDrop(aEvent, aDropData, aSession) {
    _logger.debug("personcardDNDObserver.onDrop()");
    aEvent.stopPropagation();
    try {
      // Get transferable from session
      var transferable = CC["@flock.com/rich-dnd-service;1"]
                         .getService(CI.flockIRichDNDService)
                         .getRichSelection(aSession,
                                           aDropData.flavour.contentType,
                                           aDropData.data);

      if (!transferable) {
        return;
      }

      // We know that aEvent.target is a friendcard
      var friendURN = aEvent.target.id;
      var identity = _coop.get(friendURN);
      var accountURN = identity.getParent().account.id();
      var serviceId = identity.serviceId;

      // Get appropriate sharing action spec and perform it
      var actionSpec = CC[serviceId]
                       .getService(CI.flockIWebService)
                       .getAccount(accountURN)
                       .QueryInterface(CI.flockISocialWebServiceAccount)
                       .getSharingAction(friendURN, transferable);
      _logger.debug("actionSpec: " + actionSpec);
      performActionSpec(actionSpec, null, friendURN, transferable);

      // Select the friend item
      document.getElementById("friendList").selectedItem = aEvent.target;
    } catch (ex) {
      // While we work on getRichSelection() and getSharingAction() ...
      _logger.debug("KABOOM = " + ex);
    }
  },

  getSupportedFlavours: function pc_getSupportedFlavours() {
    _logger.debug("personcardDNDObserver.getSupportedFlavours()");
    var flavors = new FlavourSet();
    flavors.appendFlavour(FLOCK_RICH_CONTENT_FLAVOUR);
    flavors.appendFlavour("text/x-moz-url");
    flavors.appendFlavour("text/unicode");
    flavors.appendFlavour("text/html");
    return flavors;
  }
}

/**
 * The following function is referenced in the mecard binding. It is the
 * onblur handler for the status textbox. Having the function reside in the
 * binding and referring to it as the handler resulted in "method is
 * undefined" at runtime.
 */
function doBlur(aEvent) {
  var inst = document.getBindingParent(aEvent.target);
  inst = document.getBindingParent(inst);
  var statusbox = document.getAnonymousElementByAttribute(inst,
                                                          "anonid",
                                                          "mecard-change-status");
  var orig = inst.getAttribute("statusMessage");
  if (statusbox.value != orig) {
    inst.saveStatus();
  }
  inst.disengageStatusChange();
  return true;
}

window.addEventListener("load", onLoad, false);
window.addEventListener("unload", unLoad, false);

/** Drag and drop code cloned from /shelf/content/shelf.js and modified */
const NS_XHTML = "http://www.w3.org/1999/xhtml";
const MARK_SELECTION_START = "\u200B\u200B\u200B\u200B\u200B";
const MARK_SELECTION_END = "\u200C\u200C\u200C\u200C\u200C";

function flock_getMarkedUpSelection(aSession) {
  var topWindow = _wm.getMostRecentWindow("navigator:browser");
  var data = "";
  var node = aSession.sourceNode;
  var serializer = new XMLSerializer();
  var parser = new DOMParser();

  var flavourset = new FlavourSet();
  flavourset.appendFlavour("text/html");
  flavourset.appendFlavour("text/unicode");
  flavourset.appendFlavour("text/x-moz-url");

  // For dragging from outside the browser (not supported)
  if (!node) {
    return null;
  }

  var contentEl = topWindow.document.getElementById("content");
  var webPageUrl = contentEl.contentDocument.URL;
  var webPageTitle = contentEl.contentTitle;

  if (webPageTitle.length == 0) {
    webPageTitle = location;
  }
  if (webPageUrl && webPageUrl.match(/^chrome:\/\/.+/)) {
    webPageUrl = "";
  }

  var contentDetector = new nsContentDetector();
  contentDetector.setTarget(node);

  // Images
  if (contentDetector.onImage) {
    var imageTitle = node.alt;
    imageURL = contentDetector.imageURL;
    // Don't insert a stupid image containing Javascript
    if (imageURL.match("javascript:")) {
      return null;
    }
    // Bug 3756, when the image alone is displayed in the browser
    if (imageURL == webPageUrl) {
      imageTitle = imageTitle.split("/").pop();
    }

    var content = "<a href='" + webPageUrl + "' title='" + imageTitle + "'>"
                + "<img border='0' alt='" + imageTitle + "' src='" + imageURL
                + "'/></a>";
    return {
      type: "image",
      title: imageTitle,
      URL: imageURL,
      content: content
    };
  }

  _logger.debug("node.nodeName: " + node.nodeName);
  switch (node.nodeName.toLowerCase()) {
    // Big hack. This is for when somebody drags the favicon onto the shelf.
    // 'image' is actually a XUL image (the favicon), while HTML images are
    // 'img'
    case "image":
      _logger.debug("node.src = " + node.getAttribute("src"));
      return {
        type: "link",
        title: webPageTitle,
        URL: webPageUrl,
        content: "<a href='" + webPageUrl + "'>" + webPageTitle + "</a>"
      };

    // Mediabar photos
    case "photo":
      header = node.title;
      header = header.replace(/%20/g, " ");
      label = node.src;
      webPageUrl = node.webPageUrl;
      return {
        type: "image",
        title: header,
        URL: webPageUrl,
        content: node.buildHTML()
      };

    // Browser tabs
    case "tab":
    case "xul:tab":
      var loc = window.top.content.document.location.toString();
      return {
        type: "link",
        title: node.label,
        URL: loc,
        content: "<a href='" + loc + "'/>"+ node.label +"</a>"
      };

    case "shelficon":
    case "richtreefolder":
      return { dragfrom: node.id };

    default:
      // Handle selections
      var selection = getSelection();

      if (selection.toString().length > 0) {
        var selection_formatted = select_src(selection);
        _logger.debug("selection_formatted.html: " + selection_formatted.html);
        data = selection_formatted.html;

        header = webPageTitle;
        label = selection;
        return {
          type: "document",
          title: webPageTitle,
          URL: webPageUrl,
          content: data
        };
      }

      if (node.nodeName.toString().toLowerCase() == "#text") {
        node = node.parentNode;
      }

      // If it is an img, make sure the src attribute is the full path
      if (node.src) {
        node.setAttribute("src", node.src);
        label = node.src;
        header = node.src;
        if (node.getAttribute("alt")) {
          header = node.getAttribute("alt");
        }
        return {
          type: "image",
          title: header,
          URL: node.src,
          content: "<a href='" + webPageUrl + "' title='" + header + "'>" +
                   "<img border='0' alt='" + header + "' src='" + node.src + "'/></a>"
        };
      }

      // If it is a link make sure the href attribute is the full path
      if (node.href) {
        node.setAttribute("href", node.href);
        label = node.href;
        header = node.textContent;
        return {
          type: "link",
          title: label,
          URL: node.href,
          content: serializer.serializeToString(node)
        };
      }
      break;
  }

  // sourceNode didn't work, let's try with dragData
  var dragData = nsTransferable.get(flavourset,
                                    nsDragAndDrop.getDragData,
                                    true)
                               .first.first;
  _logger.debug ("Flavour: " + dragData.flavour.contentType);

  switch (dragData.flavour.contentType) {
    case "text/unicode":
      return {
        type: "document",
        title: dragData.data,
        URL: "",
        content: dragData.data
      };

    case "text/html":
      return {
        type: "document",
        title: dragData.data,
        URL: "",
        content: dragData.data
      };
  }

  // Couldn't find anything
  _logger.debug("Couldn't find anything!");
  return null;
}

function getSelection() {
  var focusedWindow = document.commandDispatcher.focusedWindow;
  var winWrapper = new XPCNativeWrapper(focusedWindow,
                                        "document",
                                        "getSelection()");
  return winWrapper.getSelection();
}

function select_src(aSelection) {
  var range = aSelection.getRangeAt(0);
  var ancestorContainer = range.commonAncestorContainer;
  var doc = ancestorContainer.ownerDocument;

  var startContainer = range.startContainer;
  var endContainer = range.endContainer;
  var startOffset = range.startOffset;
  var endOffset = range.endOffset;

  // let the ancestor be an element
  if (ancestorContainer.nodeType == Node.TEXT_NODE ||
      ancestorContainer.nodeType == Node.CDATA_SECTION_NODE)
  {
    ancestorContainer = ancestorContainer.parentNode;
  }

  // for selectAll, let's use the entire document, including <html>...</html>
  // @see DocumentViewerImpl::SelectAll() for how selectAll is implemented
  try {
    if (ancestorContainer == doc.body) {
      ancestorContainer = doc.documentElement;
    }
  } catch (ex) {
  }

  // each path is a "child sequence" (a.k.a. "tumbler") that
  // descends from the ancestor down to the boundary point
  var startPath = getPath(ancestorContainer, startContainer);
  var endPath = getPath(ancestorContainer, endContainer);

  // clone the fragment of interest and reset everything to be relative to it
  // note: it is with the clone that we operate from now on
  ancestorContainer = ancestorContainer.cloneNode(true);
  startContainer = ancestorContainer;
  endContainer = ancestorContainer;
  var i;
  for (i = startPath ? startPath.length-1 : -1; i >= 0; i--) {
    startContainer = startContainer.childNodes.item(startPath[i]);
  }
  for (i = endPath ? endPath.length-1 : -1; i >= 0; i--) {
    endContainer = endContainer.childNodes.item(endPath[i]);
  }

  // add special markers to record the extent of the selection
  // note: |startOffset| and |endOffset| are interpreted either as offsets in
  // the text data or as child indices (see the Range spec)
  // (here, munging the end point first to keep the start point safe...)
  var tmpNode;
  if (endContainer.nodeType == Node.TEXT_NODE ||
      endContainer.nodeType == Node.CDATA_SECTION_NODE)
  {
    // Do some tweaks to try to avoid the view-source output to look like
    // ...<tag>]... or ...]</tag>... (where ']' marks the end of selection).
    // To get a neat output, the idea here is to remap the end point from:
    // 1. ...<tag>]...   to   ...]<tag>...
    // 2. ...]</tag>...  to   ...</tag>]...
    if ( (endOffset > 0 && endOffset < endContainer.data.length) ||
         !endContainer.parentNode ||
         !endContainer.parentNode.parentNode )
    {
      endContainer.insertData(endOffset, MARK_SELECTION_END);
    } else {
      tmpNode = doc.createTextNode(MARK_SELECTION_END);
      endContainer = endContainer.parentNode;
      if (endOffset == 0) {
        endContainer.parentNode.insertBefore(tmpNode, endContainer);
      } else {
        endContainer.parentNode.insertBefore(tmpNode,
                                             endContainer.nextSibling);
      }
    }
  } else {
    tmpNode = doc.createTextNode(MARK_SELECTION_END);
    endContainer.insertBefore(tmpNode,
                              endContainer.childNodes.item(endOffset));
  }

  if (startContainer.nodeType == Node.TEXT_NODE ||
      startContainer.nodeType == Node.CDATA_SECTION_NODE)
  {
    // Do some tweaks to try to avoid the view-source output to look like
    // ...<tag>[... or ...[</tag>... (where '[' marks the start of selection).
    // To get a neat output, the idea here is to remap the start point from:
    // 1. ...<tag>[...   to   ...[<tag>...
    // 2. ...[</tag>...  to   ...</tag>[...
    if ((startOffset > 0 && startOffset < startContainer.data.length) ||
        !startContainer.parentNode || !startContainer.parentNode.parentNode ||
        startContainer != startContainer.parentNode.lastChild)
    {
      startContainer.insertData(startOffset, MARK_SELECTION_START);
    } else {
      tmpNode = doc.createTextNode(MARK_SELECTION_START);
      startContainer = startContainer.parentNode;
      if (startOffset == 0) {
        startContainer.parentNode.insertBefore(tmpNode, startContainer);
      } else {
        startContainer.parentNode.insertBefore(tmpNode,
                                               startContainer.nextSibling);
      }
    }
  } else {
    tmpNode = doc.createTextNode(MARK_SELECTION_START);
    startContainer.insertBefore(tmpNode,
                                startContainer.childNodes.item(startOffset));
  }

  // Make relative links absolute
  makeRelativeLinksAbsolute(doc.location, ancestorContainer);

  // now extract and display the syntax highlighted source
  tmpNode = doc.createElementNS(NS_XHTML, "div");
  tmpNode.appendChild(ancestorContainer);

  var html = tmpNode.innerHTML;
  var txt = tmpNode.textContent;
  // Because there are five characters in that const.
  var html_start = html.indexOf(MARK_SELECTION_START) + 5;
  var html_end = html.indexOf(MARK_SELECTION_END);
  html = html.substring(html_start, html_end);
  // Because there are five characters in that const.
  var txt_start = txt.indexOf(MARK_SELECTION_START) + 5;
  var txt_end = txt.indexOf(MARK_SELECTION_END);
  txt = txt.substring(txt_start, txt_end);
  _logger.debug("START: " + html_start + "\n" +
                "END: " + html_end + "\n" +
                "VAL: --" + html + "--");
  _logger.debug("TXT: " + txt);
  _logger.debug("HTML: " + html);
  return { html: html, txt: txt };
}

// XXX daryl: Also ripped from viewPartialSource.js. Probably a good
// idea to keep this here, in case it changes in the core source.
function getPath(aAncestor, aNode) {
  var n = aNode;
  var p = n.parentNode;
  if (n == aAncestor || !p) {
    return null;
  }
  var path = [];
  if (!path) {
    return null;
  }
  do {
    for (var i = 0; i < p.childNodes.length; i++) {
      if (p.childNodes.item(i) == n) {
        path.push(i);
        break;
      }
    }
    n = p;
    p = n.parentNode;
  } while (n != aAncestor && p);
  return path;
}

function makeRelativeLinksAbsolute(aBaseUrl, aDOMNode) {
  // This is a recursive method: bail if no tag name
  if (!aDOMNode.tagName) {
    return;
  }

  switch (aDOMNode.tagName.toLowerCase()) {
    case "img":
      var url = aDOMNode.getAttribute("src");
      if (!url.match("http")) {
        aDOMNode.setAttribute("src", makeURLAbsolute(aBaseUrl, url));
      }
      break;

    case "a":
      var url = aDOMNode.getAttribute("href");
      if (!url.match("http")) {
        aDOMNode.setAttribute("href", makeURLAbsolute(aBaseUrl, url));
      }
      break;

    default:
      break
  }

  if (aDOMNode.hasChildNodes()) {
    var kids = aDOMNode.childNodes;
    for (var i = 0; i < kids.length; i++) {
      makeRelativeLinksAbsolute(aBaseUrl, kids[i]);
    }
  }
}

// From browser.js, required for mozextras.js
function makeURLAbsolute(aBase, aUrl) {
  try {
    var ios = CC["@mozilla.org/network/io-service;1"]
              .getService(CI.nsIIOService);
    var baseURI = ios.newURI(aBase, null, null);

    return ios.newURI(baseURI.resolve(aUrl), null, null).spec;
  } catch (ex) {
    // Special URLs throw exceptions, in this case we just return it "as is"
    return aUrl;
  }
}
