//
// BEGIN FLOCK GPL
//
// Copyright Flock Inc. 2005-2008
// http://flock.com
//
// This file may be used under the terms of the
// GNU General Public License Version 2 or later (the "GPL"),
// http://www.gnu.org/licenses/gpl.html
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
// for the specific language governing rights and limitations under the
// License.
//
// END FLOCK GPL
//

//@line 20 "/home/build/tinderbuild_blue/src/flock/mozilla/flock/base/preferences/content/searching.js"
var Cc = Components.classes;
var Ci = Components.interfaces;
var Cr = Components.results;
//@line 24 "/home/build/tinderbuild_blue/src/flock/mozilla/flock/base/preferences/content/searching.js"

const EXCLUDED_ENGINES_PREF = "flock.search.excludedEngines";
const SELECTED_SEARCH_ENGINE = "browser.search.selectedEngine";
const ENABLED_LIVE_SEARCH_ENGINES = "flock.liveSearch.enabledServices";
const LIVESEARCH_ORDERED_SVCS = "flock.liveSearch.orderedServices";
const SEARCH_SERVICE_CONTRACTID = "@mozilla.org/browser/search-service;1";
const FLOCK_SEARCH_MANAGER_CONTRACTID = "@flock.com/search-manager;1";
const OPEN_LINK_PREF = "browser.link.open_external";

const PLACEHOLDER_ICON = "chrome://flock/skin/search/placeholderIcon16.png";

var prefService = Cc["@mozilla.org/preferences-service;1"].getService(Ci.nsIPrefService).getBranch(null);
var browserSearchSvc = Cc[SEARCH_SERVICE_CONTRACTID].getService(Ci.nsIBrowserSearchService);

var gSearchPane = {

  init: function() {
    window.addEventListener("unload", this, false);
    var prefBranch = Cc["@mozilla.org/preferences-service;1"].
      getService(Ci.nsIPrefBranch2);
    prefBranch.addObserver(SELECTED_SEARCH_ENGINE, this, false);
    gLiveSearchView = new LiveSearchView(document.getElementById("liveSearchServices"));
    setupLiveSearchList();
    buildDefaultSearchPopup();
    buildSearchElsewhere();
    flock_setupSearchPrefs();
  },
  
  uninit: function() {
    var prefBranch = Cc["@mozilla.org/preferences-service;1"].
      getService(Ci.nsIPrefBranch2);
    prefBranch.removeObserver(SELECTED_SEARCH_ENGINE, this);
    window.removeEventListener("unload", this, false);
  },

  handleEvent: function(aEvent) {
    if (aEvent.type == "unload") {
      this.uninit();
    }
  },

  observe: function(aSubject, aTopic, aData) {
    if (aTopic == "nsPref:changed") {
      switch (aData) {
        case SELECTED_SEARCH_ENGINE:
          flock_setDefaultEnginePopup();
          break;
      }
    }
  }

};

var gLiveSearchView;

function flock_getLocalizedPref(aPref, aDefault) {
  try {
    return prefService.getComplexValue(aPref, Ci.nsIPrefLocalizedString).data;
  }
  catch (ex) {
  }

  try {
    return prefService.getCharPref(aPref);
  }
  catch (ex) {
  }

  return aDefault;
}

function flock_setLocalizedPref(aPref, aValue) {
  try {
    var pls = Cc["@mozilla.org/pref-localizedstring;1"]
              .createInstance(Ci.nsIPrefLocalizedString);
    pls.data = aValue;
    prefService.setComplexValue(aPref, Ci.nsIPrefLocalizedString, pls);
  }
  catch (ex) {
  }
}

function flock_setupSearchPrefs() {
  var flockSearchListbox = document.getElementById("flockSearchListbox");
  var engineItems = flockSearchListbox.getElementsByTagName("listitem");
  var excludedEngines = flock_getLocalizedPref(EXCLUDED_ENGINES_PREF, "");
  for (var i =0; i < engineItems.length; i++) {
    if (excludedEngines.indexOf(engineItems[i].childNodes[1].getAttribute("label")) == -1) {
      engineItems[i].firstChild.firstChild.setAttribute("checked", true);
    }
  }
  flock_setDefaultEnginePopup();
}

function flock_saveSelectedEnginePrefs() {
  var flockSearchListbox = document.getElementById("flockSearchListbox");
  var engineItems = flockSearchListbox.getElementsByTagName("listitem");
  flock_setLocalizedPref(EXCLUDED_ENGINES_PREF, "");
  var prefString = "";
  for (var i = 0; i < engineItems.length; i++) {
    if (engineItems[i].firstChild.firstChild.checked == false) {
      prefString = prefString + engineItems[i].childNodes[1].getAttribute("label") + ",";
    }
  }
  flock_setLocalizedPref(EXCLUDED_ENGINES_PREF, prefString);
}

function flock_searchUpdatePref(aTarget) {
  flock_setBoolPref(aTarget.getAttribute("id"), aTarget.checked);
}

function flock_setDefaultEngine() {
  var selectedEngine = $("flockDefaultSearchMenu").selectedItem;
  var engineName = selectedEngine.getAttribute("label");

  var currentEngineName = browserSearchSvc.currentEngine.name;

  var engines = browserSearchSvc.getEngines({});
  for each (var engine in engines) {
    if (engine.name == engineName) {
      browserSearchSvc.currentEngine = engine;
      break;
    }
  }

  // User has selected a different search engine, so we no longer want
  // it to appear in the "Search Elsewhere" list. This means we need to
  // ADD it to the list of EXCLUDED search engines.
  // Also, we want to REMOVE the previous search engine, so it once again
  // (or for first time) appears under "Search Elsewhere".
  var oldList = flock_getLocalizedPref(EXCLUDED_ENGINES_PREF, "").split(",");
  var newList = [];
  newList.push(engineName);
  for each (var engine in oldList) {
    if (engine != currentEngineName && engine != engineName) {
      newList.push(engine);
    }
  }
  flock_setLocalizedPref(EXCLUDED_ENGINES_PREF, newList.join(","));

  flock_setupSearchPrefs();
}

function flock_setDefaultEnginePopup() {
  // Has the user set this preference?
  var userPrefSet = prefService.prefHasUserValue(SELECTED_SEARCH_ENGINE);

  // Did we find the engine set in the preferences in the menulist?
  var prefMatches = false;

  if (userPrefSet) {
    var currentEngine = browserSearchSvc.currentEngine.name;
    setMenulistSelection(currentEngine);
  }

  function setMenulistSelection(aEngine) {
    var flockSearchListbox = document.getElementById("flockSearchListbox");
    var engineItems = flockSearchListbox.getElementsByTagName("listitem");
    for (var i = 0; i < engineItems.length; i++) {
      if (engineItems[i].childNodes[1].getAttribute("label") == aEngine) {
        $("flockDefaultSearchMenu").selectedIndex = i;
        prefMatches = true;
      }
    }
  }

  // If we didn't find it, remove the invalid value so we revert to default.
  if (!prefMatches) {
    var defaultEngine = browserSearchSvc.defaultEngine.name;
    setMenulistSelection(defaultEngine);
    if (userPrefSet) {
      prefService.clearUserPref(SELECTED_SEARCH_ENGINE);
    }
  }
}


function setupLiveSearchList() {
  var searchMgrSvc = Cc[FLOCK_SEARCH_MANAGER_CONTRACTID].getService(Ci.flockISearchManager);
  var liveSearchListEnum = searchMgrSvc.orderedServices;
  var liveSearchContainer = document.getElementById("liveSearchServices");
  var service, listitem, listcell, checkbox;
  while (liveSearchListEnum.hasMoreElements()) {
    service = liveSearchListEnum.getNext();
    listitem = document.createElement("listitem");
    listitem.setAttribute("value", service.shortName);
    listitem.setAttribute("allowevents", true);
    
    listcell = document.createElement("listcell");
    checkbox = document.createElement("checkbox");
    checkbox.setAttribute("checked", searchMgrSvc.isServiceEnabled(service.shortName));
    checkbox.setAttribute("oncommand", "gLiveSearchView.synchToPref();");
    listcell.appendChild(checkbox);
    listitem.appendChild(listcell);
    
    listcell = document.createElement("listcell");
    listcell.setAttribute("label", service.serviceName);
    listitem.appendChild(listcell);
    
    liveSearchContainer.appendChild(listitem);
  }
}

var LiveSearchView = function(aListbox) {
  this.mListbox = aListbox;
  this.mSelectedIndex = null;
  this.mTotalItems = null;
  this.mGoDownButton = null;
  this.mGoUpButton = null;
  
  this.init();
}

LiveSearchView.prototype = {
  init: function() {
    this.mGoDownButton = document.getElementById("goDownButton");
    this.mGoUpButton = document.getElementById("goUpButton");
    this.mGoUpButton.setAttribute("disabled", true);
    this.mGoDownButton.setAttribute("disabled", true);
  },
  
  onSelect: function() {
    // buttons only work if an engine is selected and it's not the last engine
    var disableButtons = (this.mListbox.selectedIndex == -1);
    var lastSelected = (this.mListbox.selectedIndex == this.mListbox.getRowCount() - 1);
    var firstSelected = (this.mListbox.selectedIndex == 0);

    this.mGoUpButton.setAttribute("disabled", disableButtons || firstSelected);
    this.mGoDownButton.setAttribute("disabled", disableButtons || lastSelected);
  },
  
  bump: function(aDir) {
    //liveSearchOrder.mListbox.selectedIndex--;
    var lastIndex = this.mListbox.selectedIndex;
    var clonedNode = this.mListbox.selectedItem.cloneNode(true);
    clonedNode.removeAttribute("selected");
    clonedNode.removeAttribute("current");
    
    if (aDir == 1) {
      this.mListbox.insertBefore(clonedNode, this.mListbox.selectedItem.previousSibling);
      this.mListbox.removeItemAt(lastIndex+1);
    } else if (aDir == -1) {
      var nextSibling = this.mListbox.selectedItem.nextSibling;
      this.mListbox.removeItemAt(lastIndex);
      if (nextSibling.nextSibling) {
        this.mListbox.insertBefore(clonedNode, nextSibling.nextSibling);
      } else {
        this.mListbox.appendChild(clonedNode);
      }
    }
    
    this.mListbox.selectedItem = clonedNode;
    this.mListbox.ensureElementIsVisible(this.mListbox.selectedItem); 
    this.synchToPref();
    this.onSelect();
  },
  
  synchToPref: function() {
    var liveSearchItems = this.mListbox.childNodes;
    var enabledString = "";
    var orderedString = "";
    for (var i = 0; i < liveSearchItems.length; i++) {
      var shortName = liveSearchItems[i].getAttribute("value");
      if (!shortName || !shortName.length) {
        continue;
      }
      orderedString += shortName + ",";
      if (liveSearchItems[i].firstChild.firstChild.getAttribute("checked") == "true") {
        enabledString += shortName + ",";
      }
    }
    prefService.setCharPref(ENABLED_LIVE_SEARCH_ENGINES, enabledString);
    prefService.setCharPref(LIVESEARCH_ORDERED_SVCS, orderedString);
  }
};
  
function getOpenLinkPref() {
  var openLinkPref = prefService.getIntPref(OPEN_LINK_PREF);
  if (openLinkPref == "3") {
    return "tab";
  } else {
    return "window";
  }
}

function buildDefaultSearchPopup() {
  var engines = browserSearchSvc.getEngines({});
  var defaultSearchPopup = document.getElementById("flockDefaultSearchPopup");
  for (var i = 0; i < engines.length; i++ ) {
    var menuItem = document.createElement("menuitem");
    menuItem.setAttribute("label", engines[i].name);
    menuItem.setAttribute("image", engines[i].iconURI ? engines[i].iconURI.spec
                                                      : PLACEHOLDER_ICON);
    menuItem.setAttribute("class", "menuitem-iconic");
    menuItem.setAttribute("oncommand", "flock_setDefaultEngine();");
    defaultSearchPopup.appendChild(menuItem);
  }
}

function buildSearchElsewhere() {
  var engines = browserSearchSvc.getEngines({});
  var searchListBox = document.getElementById("flockSearchListbox");
  var listitem, listcell, checkbox;
  for (var i = 0; i < engines.length; i++ ) {
    listitem = document.createElement("listitem");
    listitem.setAttribute("allowevents", true);
    
    listcell = document.createElement("listcell");
    checkbox = document.createElement("checkbox");
    checkbox.setAttribute("oncommand", "flock_saveSelectedEnginePrefs();");
    listcell.appendChild(checkbox);
    listitem.appendChild(listcell);
    
    listcell = document.createElement("listcell");
    listcell.setAttribute("label", engines[i].name);
    listcell.setAttribute("image", engines[i].iconURI ? engines[i].iconURI.spec
                                                      : PLACEHOLDER_ICON);
    listcell.setAttribute("class", "listcell-iconic");
    listitem.appendChild(listcell);
    
    searchListBox.appendChild(listitem);
  }
}
